==> apracjs-13.1.29.tgz: package/lib/handlers/gspx/gspx.js <==
﻿
/*!
 * apracjs
 * Copyright(c) 2017 by wssf
 * 2017-04-26
 * 处理gspx页面的请求
 */

'use strict'
var fs = require('fs');
var path = require('path');
var debug = require('debug')('aprac:gspx');
var Handler = require('../handler.js');
var PageTag = require('./pagetag.js');
var Context = require('./context.js');
var config = require('../../config.js').apracConfig;
var parserCache = require('./parserCache.js');
var sysConsts = require('../../scripts/sysConsts.json');
var loginPage = require('../../scripts/loginPageInfo.js');
var utils = require('../../common/utils.js');
var htmlCache = {}; // 没有实现jsx的页面作为html输出，缓存起来

class GspxHandler extends Handler {
  constructor(app, distributor) {
    super(app, '.gspx', distributor);
  }

  enableSession(req) {
    return true;
  }

  processRequest(req, res, done) {
    super.processRequest(req, res, done);

    var httpContext = new Context(req, res);
    this.invokeController(httpContext, (result) => {
      if (result !== true) {
        if (result == undefined) {
          return res.end('{}'); // 如果jsx实现了init，则一律视为api服务，如果业务init忘了end，这里必须强制结束，不然客户端无法收到响应；
        }

        if (typeof result == "object" && result.responseRedirect) { // init返回重定向url
          res.writeHead(302, { 'Location': result.responseRedirect });
          return res.end();
        }

        if (result === false) {
          return; // init返回false，不再继续处理页面
        }

        if (typeof result === 'function') {
          return result(res); // init返回cb，处理init里面的回调
        }

        return this.endResponse(res, result.message); // 处理init或context的异常信息
      }

      if (typeof httpContext.__contextData != "function") {
        return this.renderPage(httpContext);
      }

      // 继续处理gspx页面
      var RenderPage = utils.createDelegate(this, this.renderPage);

      var callback = function (json, err) {
        httpContext.__contextData = json;
        RenderPage(httpContext, err);
      };

      // 有些业务的context使用了async,但却没有try..catch，当服务端出现异常时前端一直转圈没有消息提示
      var contextFun = httpContext.__contextData;
      if (Object.prototype.toString.call(contextFun) == '[object AsyncFunction]') {
        async function processOn(cb) {
          try {
            await contextFun(cb);
          } catch (e) {
            callback(null, e);
          }
        }
        processOn(callback);
      } else {
        httpContext.__contextData(callback);
      }
    });
  }

  invokeController(httpContext, done) {
    this._distrributor ? this.distributorInvokeController(httpContext, done) : this.selfInvokeController(httpContext, done); //加入分拨器
  }

  distributorInvokeController(httpContext, done) {
    var req = httpContext.req;
    var filename = req.parsed.filename;
    var needFile = false;
    // if (!parserCache.isCached(filename) && !fs.existsSync(filename)) {//通过判断是否已经解析缓存了，来决定是否再获取gspx
    //   needFile = true;
    // }
    // needFile = false;
    var data = {
      filename: req.parsed.relativepath,
      needFile: needFile
    }
    this._distrributor.get(req, data, "gspx", (getData) => {
      if (getData.redirectUrl && getData.redirectUrl.length > 0) {//是否重定向
        done({ responseRedirect: getData.redirectUrl });
        return;
      }
      // if(needFile){
      //   //直接下载对应文件,并保持到对应目录下
      //   this._distrributor.down(req, getData.url);
      // }
      httpContext.__contextData = getData.context;//上下文内容
      httpContext.pageInfo.setControler({
        WebMethod: ''
      });//指定是否有远程方法调用
      done(true);
    });
  }

  _getPageClass(req, done) {
    var filename = req.parsed.filename;
    var filedir = req.parsed.filedir;

    var jsxPath = path.join(filedir, path.basename(filename, '.gspx') + '.jsx');
    var PageClass = null;
    var jsxCache = require.cache[jsxPath]; // 直接检查底层的缓存，不用自己单独建立缓存管理
    if (!jsxCache) {
      if (htmlCache[jsxPath]) {
        done(true);
        return null;
      }

      if (!fs.existsSync(jsxPath)) {
        htmlCache[jsxPath] = true;
        done(true);
        return null; // 没有文件，说明没有实现服务端业务，gspx作为html处理, 缓存目的是避免重复existsSync
      }

      var stats = fs.statSync(jsxPath); // 第一次加载jsx，读取最后更新时间，输出到浏览器端，方便排查更新发版问题
     

      PageClass = require(jsxPath);
      PageClass.lastModified = stats.mtime;
    } else {
      PageClass = jsxCache.exports; // 如果有缓存，就不用在检查文件是否存在，避免阻塞，提高并发效率
    }

    req.res.lastModified = PageClass.lastModified;
    return PageClass;
  }

  async selfInvokeController(httpContext, done) {
    let req = httpContext.req;
    var answer = undefined;
    let PageClass;

    try {
      PageClass = this._getPageClass(req, done); // js解析可能出现语法异常
      if (!PageClass) return;

      answer = await loginPage.checkHasLogin(req, typeof (PageClass.prototype.needLogin) == "function");
    }
    catch (t) {
      var checkInfo = {};
      if (t.responseRedirect) {
        req.parsed.pathname = t.responseRedirect; // 重定向到固定目录
        req.url = t.responseRedirect;
        return this.processRequest(req, req.res, done);
      }
      else if (t.stack || t.message) { // 捕获抛出的异常
        checkInfo.message = t.message;
        return done(checkInfo);
      }
      else {
        checkInfo.message = t.resMsg ? t.resMsg : t; // 业务写死的resMsg或者throw的字符串
        return done(checkInfo);
      }
    }

    if (!answer) {
      var checkInfo = {};
      if (!httpContext.getIsAsyncRequest()) {
        checkInfo.responseRedirect = `${loginPage.loginPageUrl}?Redir=${req.parsed.pathname}`;
      }
      else {
        checkInfo.message = sysConsts.Session_TimeoutError;
      }
      return done(checkInfo);
    }

    var controller = new PageClass();
    controller.httpContext = httpContext; // 注入依赖
    controller.WebMethod = PageClass.WebMethod;
    controller.req = req;
    controller.res = req.res;
    controller.user = loginPage.userInfo(req);
    controller.pageParams = httpContext.pageParams;
    utils.copyFunTo(Context.prototype, controller, httpContext); // 注入依赖，直接访问更便捷些

    try {
      var interceptor = PageClass.Interceptor ? new PageClass.Interceptor(controller.req) : null;
      if (typeof (controller.init) == "function") {
        var result = await this.wrapInterceptor(interceptor, "init", controller, [req.res]);               // 页面初始化
        if (controller.redirectUrl) {
          return done({ responseRedirect: controller.redirectUrl });  // 页面已经采取重定向，不继续处理

        }
        return done(result);   // 如果页面实现了init，则视为已经采取自定义输出了，结束页面请求，不继续处理gspx了
      }

      if (typeof (controller.context) == "function") {
        httpContext.__contextData = await this.wrapInterceptor(interceptor, "context", controller);// 执行context()读取json当前上下文
        if (httpContext.__contextData && httpContext.__contextData.responseRedirect) {
          return done({ responseRedirect: httpContext.__contextData.responseRedirect });// 页面已经采取重定向，不继续处理
        }
      }
    }
    catch (ex) {
      return done(ex);
    }
    finally {
    }

    httpContext.pageInfo.setControler(controller);
    done(true); // 只有返回true才继续执行
  }

  async renderPage(httpContext, err) {
    var res = httpContext.res;
    var req = httpContext.req;
    var done = req.done;

    if (err) {
      return this.endResponse(res, utils.makeExceptionHtml(err)); // 打出错误信息
    }

    var filename = req.parsed.filename;
    var gspxCache = parserCache.getGspxCache(filename);
    if (!gspxCache) {
      debug('页面解析错误：%s', filename);
      return done();
    }

    var tag = new PageTag(this, gspxCache.json, httpContext);
    await gspxCache.initContextSession(tag);
    tag.renderPage();
  }

  endResponse(res, html) {
    res.setHeader('Content-Type', "text/html; charset=utf-8");
    res.setHeader('Cache-Control', 'no-cache');
    if (res.lastModified) res.setHeader('X-jsx-lastModified', res.lastModified);
    res.end(html);
  }
}

module.exports = GspxHandler;

==> apracjs-15.1.3.tgz: package/lib/handlers/gspx/gspx.js <==
﻿
/*!
 * apracjs
 * Copyright(c) 2017 by wssf
 * 2017-04-26
 * 处理gspx页面的请求
 */

'use strict'
var fs = require('fs');
var path = require('path');
var debug = require('debug')('aprac:gspx');
var Handler = require('../handler.js');
var PageTag = require('./pagetag.js');
var Context = require('./context.js');
var config = require('../../config.js').apracConfig;
var parserCache = require('./parserCache.js');
var sysConsts = require('../../scripts/sysConsts.json');
var loginPage = require('../../scripts/loginPageInfo.js');
var utils = require('../../common/utils.js');
var htmlCache = {}; // 没有实现jsx的页面作为html输出，缓存起来

class GspxHandler extends Handler {
  constructor(app, distributor) {
    super(app, '.gspx', distributor);
  }

  enableSession(req) {
    return true;
  }

  processRequest(req, res, done) {
    super.processRequest(req, res, done);

    var httpContext = new Context(req, res);
    this.invokeController(httpContext, (result) => {
      if (result !== true) {
        if (result == undefined) {
          return res.end('{}'); // 如果jsx实现了init，则一律视为api服务，如果业务init忘了end，这里必须强制结束，不然客户端无法收到响应；
        }

        if (typeof result == "object" && result.responseRedirect) { // init返回重定向url
          res.writeHead(302, { 'Location': result.responseRedirect });
          return res.end();
        }

        if (result === false) {
          return; // init返回false，不再继续处理页面
        }

        if (typeof result === 'function') {
          return result(res); // init返回cb，处理init里面的回调
        }

        return this.endResponse(res, result.message); // 处理init或context的异常信息
      }

      if (typeof httpContext.__contextData != "function") {
        return this.renderPage(httpContext);
      }

      // 继续处理gspx页面
      var RenderPage = utils.createDelegate(this, this.renderPage);

      var callback = function (json, err) {
        httpContext.__contextData = json;
        RenderPage(httpContext, err);
      };

      // 有些业务的context使用了async,但却没有try..catch，当服务端出现异常时前端一直转圈没有消息提示
      var contextFun = httpContext.__contextData;
      if (Object.prototype.toString.call(contextFun) == '[object AsyncFunction]') {
        async function processOn(cb) {
          try {
            await contextFun(cb);
          } catch (e) {
            callback(null, e);
          }
        }
        processOn(callback);
      } else {
        httpContext.__contextData(callback);
      }
    });
  }

  invokeController(httpContext, done) {
    this._distrributor ? this.distributorInvokeController(httpContext, done) : this.selfInvokeController(httpContext, done); //加入分拨器
  }

  distributorInvokeController(httpContext, done) {
    var req = httpContext.req;
    var filename = req.parsed.filename;
    var needFile = false;
    // if (!parserCache.isCached(filename) && !fs.existsSync(filename)) {//通过判断是否已经解析缓存了，来决定是否再获取gspx
    //   needFile = true;
    // }
    // needFile = false;
    var data = {
      filename: req.parsed.relativepath,
      needFile: needFile
    }
    this._distrributor.get(req, data, "gspx", (getData) => {
      if (getData.redirectUrl && getData.redirectUrl.length > 0) {//是否重定向
        done({ responseRedirect: getData.redirectUrl });
        return;
      }
      // if(needFile){
      //   //直接下载对应文件,并保持到对应目录下
      //   this._distrributor.down(req, getData.url);
      // }
      httpContext.__contextData = getData.context;//上下文内容
      httpContext.pageInfo.setControler({
        WebMethod: ''
      });//指定是否有远程方法调用
      done(true);
    });
  }

  _getPageClass(req, done) {
    var filename = req.parsed.filename;
    var filedir = req.parsed.filedir;

    var jsxPath = path.join(filedir, path.basename(filename, '.gspx') + '.jsx');
    var PageClass = null;
    var jsxCache = require.cache[jsxPath]; // 直接检查底层的缓存，不用自己单独建立缓存管理
    if (!jsxCache) {
      if (htmlCache[jsxPath]) {
        done(true);
        return null;
      }

      if (!fs.existsSync(jsxPath)) {
        htmlCache[jsxPath] = true;
        done(true);
        return null; // 没有文件，说明没有实现服务端业务，gspx作为html处理, 缓存目的是避免重复existsSync
      }

      var stats = fs.statSync(jsxPath); // 第一次加载jsx，读取最后更新时间，输出到浏览器端，方便排查更新发版问题
     

      PageClass = require(jsxPath);
      PageClass.lastModified = stats.mtime;
    } else {
      PageClass = jsxCache.exports; // 如果有缓存，就不用在检查文件是否存在，避免阻塞，提高并发效率
    }

    req.res.lastModified = PageClass.lastModified;
    return PageClass;
  }

  async selfInvokeController(httpContext, done) {
    let req = httpContext.req;
    var answer = undefined;
    let PageClass;

    try {
      PageClass = this._getPageClass(req, done); // js解析可能出现语法异常
      if (!PageClass) return;

      answer = await loginPage.checkHasLogin(req, typeof (PageClass.prototype.needLogin) == "function");
    }
    catch (t) {
      var checkInfo = {};
      if (t.responseRedirect) {
        req.parsed.pathname = t.responseRedirect; // 重定向到固定目录
        req.url = t.responseRedirect;
        return this.processRequest(req, req.res, done);
      }
      else if (t.stack || t.message) { // 捕获抛出的异常
        checkInfo.message = t.message;
        return done(checkInfo);
      }
      else {
        checkInfo.message = t.resMsg ? t.resMsg : t; // 业务写死的resMsg或者throw的字符串
        return done(checkInfo);
      }
    }

    if (!answer) {
      var checkInfo = {};
      if (!httpContext.getIsAsyncRequest()) {
        checkInfo.responseRedirect = `${loginPage.loginPageUrl}?Redir=${req.parsed.pathname}`;
      }
      else {
        checkInfo.message = sysConsts.Session_TimeoutError;
      }
      return done(checkInfo);
    }

    var controller = new PageClass();
    controller.httpContext = httpContext; // 注入依赖
    controller.WebMethod = PageClass.WebMethod;
    controller.req = req;
    controller.res = req.res;
    controller.user = loginPage.userInfo(req);
    controller.pageParams = httpContext.pageParams;
    utils.copyFunTo(Context.prototype, controller, httpContext); // 注入依赖，直接访问更便捷些

    try {
      var interceptor = PageClass.Interceptor ? new PageClass.Interceptor(controller.req) : null;
      if (typeof (controller.init) == "function") {
        var result = await this.wrapInterceptor(interceptor, "init", controller, [req.res]);               // 页面初始化
        if (controller.redirectUrl) {
          return done({ responseRedirect: controller.redirectUrl });  // 页面已经采取重定向，不继续处理

        }
        return done(result);   // 如果页面实现了init，则视为已经采取自定义输出了，结束页面请求，不继续处理gspx了
      }

      if (typeof (controller.context) == "function") {
        httpContext.__contextData = await this.wrapInterceptor(interceptor, "context", controller);// 执行context()读取json当前上下文
        if (httpContext.__contextData && httpContext.__contextData.responseRedirect) {
          return done({ responseRedirect: httpContext.__contextData.responseRedirect });// 页面已经采取重定向，不继续处理
        }
      }
    }
    catch (ex) {
      return done(ex);
    }
    finally {
    }

    httpContext.pageInfo.setControler(controller);
    done(true); // 只有返回true才继续执行
  }

  async renderPage(httpContext, err) {
    var res = httpContext.res;
    var req = httpContext.req;
    var done = req.done;

    if (err) {
      return this.endResponse(res, utils.makeExceptionHtml(err, null, req)); // 打出错误信息
    }

    var filename = req.parsed.filename;
    var gspxCache = parserCache.getGspxCache(filename);
    if (!gspxCache) {
      debug('页面解析错误：%s', filename);
      return done();
    }

    var tag = new PageTag(this, gspxCache.json, httpContext);
    await gspxCache.initContextSession(tag);
    tag.renderPage();
  }

  endResponse(res, html) {
    res.setHeader('Content-Type', "text/html; charset=utf-8");
    res.setHeader('Cache-Control', 'no-cache');
    if (res.lastModified) res.setHeader('X-jsx-lastModified', res.lastModified);
    res.end(html);
  }
}

module.exports = GspxHandler;
